/*
 * test_text.c
 *
 * Copyright 2022 Jonathan Schöbel <jonathan@Ubermos-2019>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 */


#include <check.h>
#include <limits.h>
#include <stdbool.h>
#include <stdlib.h>

/* lower SIZE_MAX as we try to reach it */
#include <limits.h>
#define SIZE_MAX 30

#include "macro.h"
#include "error.h"

/* C file is needed, because we want to override CHUNK_SIZE */
#define CHUNK_SIZE 5
#include "text.c"


START_TEST(test_text)
{
	struct SH_Error error;
	struct SH_Text * text;

	text = SH_Text_new (NULL);
	ck_assert_int_ne ((long int) text, (long int) NULL);

	ck_assert_int_ge  (text->data->size, CHUNK_SIZE);

	SH_Text_free (text, NULL);

	error.type = UNDEFINED;
	text = SH_Text_new (&error);
	ck_assert_int_ne ((long int) text, (long int) NULL);
	ck_assert_int_eq (error.type, SUCCESS);

	ck_assert_int_ge  (text->data->size, CHUNK_SIZE);

	error.type = UNDEFINED;
	SH_Text_free (text, &error);
	ck_assert_int_eq (error.type, SUCCESS);
}
END_TEST

START_TEST(test_text_string)
{
	struct SH_Error error;
	struct SH_Text * text;

	/* without error */
	text = SH_Text_new_from_string ("12345", NULL);
	ck_assert_int_ne ((long int) text, (long int) NULL);

	ck_assert_int_eq (text->data->length, 5);
	ck_assert_int_eq (text->data->size, 6);
	ck_assert_str_eq (text->data->text, "12345");
	ck_assert_ptr_eq (text->data->next, NULL);

	SH_Text_free (text, NULL);


	/* with error */
	error.type = UNDEFINED;
	text = SH_Text_new_from_string ("12345", &error);
	ck_assert_int_ne ((long int) text, (long int) NULL);
	ck_assert_int_eq (error.type, SUCCESS);

	ck_assert_int_eq (text->data->length, 5);
	ck_assert_int_eq (text->data->size, 6);
	ck_assert_str_eq (text->data->text, "12345");
	ck_assert_ptr_eq (text->data->next, NULL);

	SH_Text_free (text, NULL);

}
END_TEST

START_TEST (test_text_copy)
{
	struct SH_Error error;
	struct SH_Text * text;
	struct SH_Text * copy;
	char * text_string;
	char * copy_string;

	text = SH_Text_new_from_string ("Text12345", NULL);
	text_string = SH_Text_get_string (text, 0, 10, NULL, NULL);

	/* without error */
	copy = SH_Text_copy (text, NULL);
	ck_assert_ptr_ne (copy, NULL);
	ck_assert_ptr_ne (copy, text);
	copy_string = SH_Text_get_string (copy, 0, 10, NULL, NULL);
	ck_assert_str_eq (copy_string, text_string);

	free (copy_string);
	SH_Text_free (copy, NULL);

	/* with error */
	error.type = UNDEFINED;
	copy = SH_Text_copy (text, &error);
	ck_assert_int_eq (error.type, SUCCESS);
	ck_assert_ptr_ne (copy, NULL);
	ck_assert_ptr_ne (copy, text);
	copy_string = SH_Text_get_string (copy, 0, 10, NULL, NULL);
	ck_assert_str_eq (copy_string, text_string);

	free (copy_string);
	SH_Text_free (copy, NULL);

	free (text_string);
	SH_Text_free (text, NULL);
}
END_TEST

START_TEST(test_text_get_char)
{
	struct SH_Error error;
	struct SH_Text * text;
	unsigned int i;
	char * c;

	text = SH_Text_new_from_string ("abcd", NULL);
	SH_Text_append_string (text, "efgh", NULL);
	SH_Text_append_string (text, "ijkl", NULL);

	/* success - without error */
	for (i = 0; i < 12; i++)
	{
		c = SH_Text_get_char (text, i, NULL);
		ck_assert_ptr_ne (c, NULL);
		ck_assert_int_eq (*c, 97 + i);
		free (c);
	}

	/* wrong index - without error */
	c = SH_Text_get_char (text, 13, NULL);
	ck_assert_ptr_eq (c, NULL);

	/* success - with error */
	for (i = 0; i < 12; i++)
	{
		error.type = UNDEFINED;
		c = SH_Text_get_char (text, i, &error);
		ck_assert_ptr_ne (c, NULL);
		ck_assert_int_eq (*c, 97 + i);
		ck_assert_int_eq (error.type, SUCCESS);
		free (c);
	}

	/* wrong index - with error */
	error.type = UNDEFINED;
	c = SH_Text_get_char (text, 13, &error);
	ck_assert_ptr_eq (c, NULL);
	ck_assert_int_eq (error.type, VALUE_ERROR);

	SH_Text_free (text, NULL);
}
END_TEST

START_TEST(test_text_get_string)
{
	struct SH_Error error;
	struct SH_Text * text;
	char * result;
	size_t length = -1;
	char index[2] = {1, 0};

	text = SH_Text_new (NULL);
	for (; index[0] < 100; index[0]++)
	{
		SH_Text_append_string (text, &index, NULL);
	}

	/* single segment - without error */
	result = SH_Text_get_string (text, 59, 8, &length, NULL);
	ck_assert_str_eq (result, "<=>?@ABC");
	ck_assert_int_eq (length, 8);
	free (result);

	/* multiple segment - without error */
	result = SH_Text_get_string (text, 47, 43, &length, NULL);
	ck_assert_str_eq (result, "0123456789:;<=>?@"
	                          "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
	ck_assert_int_eq (length, 43);
	free (result);

	/* pass over end - without error */
	result = SH_Text_get_string (text, 96, 10, &length, NULL);
	ck_assert_str_eq (result, "abc");
	ck_assert_int_eq (length, 3);
	free (result);

	/* out of range - without error */
	result = SH_Text_get_string (text, 99, 0, &length, NULL);
	ck_assert_ptr_eq (result, NULL);
	ck_assert_int_eq (length, 0);


	/* single segment - with error */
	error.type = UNDEFINED;
	result = SH_Text_get_string (text, 59, 8, &length, &error);
	ck_assert_str_eq (result, "<=>?@ABC");
	ck_assert_int_eq (length, 8);
	ck_assert_int_eq (error.type, SUCCESS);
	free (result);

	/* multiple segment - with error */
	error.type = UNDEFINED;
	result = SH_Text_get_string (text, 47, 43, &length, &error);
	ck_assert_str_eq (result, "0123456789:;<=>?@"
	                          "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
	ck_assert_int_eq (length, 43);
	ck_assert_int_eq (error.type, SUCCESS);
	free (result);

	/* pass over end - with error */
	error.type = UNDEFINED;
	result = SH_Text_get_string (text, 96, 10, &length, &error);
	ck_assert_str_eq (result, "abc");
	ck_assert_int_eq (length, 3);
	ck_assert_int_eq (error.type, SUCCESS);
	free (result);

	/* out of range - without error */
	error.type = UNDEFINED;
	result = SH_Text_get_string (text, 99, 0, &length, &error);
	ck_assert_ptr_eq (result, NULL);
	ck_assert_int_eq (length, 0);
	ck_assert_int_eq (error.type, VALUE_ERROR);

	SH_Text_free (text, NULL);
}
END_TEST

START_TEST (test_text_append)
{
	struct SH_Error error;
	struct SH_Text * text1;
	struct SH_Text * text2;
	bool boolean;
	char * string;
	size_t length;

	text1 = SH_Text_new (NULL);
	text2 = SH_Text_new (NULL);

	/* append_string - without error */
	boolean = SH_Text_append_string (text1, "Text1", NULL);
	ck_assert_int_eq (boolean, TRUE);
	string = SH_Text_get_string (text1, 0, SIZE_MAX, &length, NULL);
	ck_assert_str_eq (string, "Text1");
	ck_assert_int_eq (length, 5);
	free (string);

	/* append_string - with error */
	error.type = UNDEFINED;
	boolean = SH_Text_append_string (text2, "Text2", &error);
	ck_assert_int_eq (boolean, TRUE);
	ck_assert_int_eq (error.type, SUCCESS);
	string = SH_Text_get_string (text2, 0, SIZE_MAX, &length, NULL);
	ck_assert_str_eq (string, "Text2");
	ck_assert_int_eq (length, 5);
	free (string);


	/* append_text - without error */
	boolean = SH_Text_append_text (text1, text2, NULL);
	ck_assert_int_eq (boolean, TRUE);
	string = SH_Text_get_string (text1, 0, SIZE_MAX, &length, NULL);
	ck_assert_str_eq (string, "Text1Text2");
	ck_assert_int_eq (length, 10);
	free (string);

	/* append_text - with error */
	error.type = UNDEFINED;
	boolean = SH_Text_append_text (text1, text2, &error);
	ck_assert_int_eq (boolean, TRUE);
	ck_assert_int_eq (error.type, SUCCESS);
	string = SH_Text_get_string (text1, 0, SIZE_MAX, &length, NULL);
	ck_assert_str_eq (string, "Text1Text2Text2");
	ck_assert_int_eq (length, 15);
	free (string);


	/* join - no error */
	SH_Text_join (text1, text2);
	string = SH_Text_get_string (text1, 0, SIZE_MAX, &length, NULL);
	ck_assert_str_eq (string, "Text1Text2Text2Text2");
	ck_assert_int_eq (length, 20);
	free (string);


	SH_Text_free (text1, NULL);
}
END_TEST

Suite * text_suite (void)
{
	Suite *s;
	TCase *tc_core;

	s = suite_create ("Testsuite SeFHT Text");

	/* Core test case */
	tc_core = tcase_create ("Core");

	tcase_add_test (tc_core, test_text);
	tcase_add_test (tc_core, test_text_string);
	tcase_add_test (tc_core, test_text_copy);
	tcase_add_test (tc_core, test_text_get_char);
	tcase_add_test (tc_core, test_text_get_string);
	tcase_add_test (tc_core, test_text_append);
	suite_add_tcase (s, tc_core);

	return s;
}

int main (int argc, char **argv)
{
	int number_failed;
	Suite *s;
	SRunner *sr;

	s = text_suite ();
	sr = srunner_create (s);

	srunner_run_all (sr, CK_NORMAL);
	number_failed = srunner_ntests_failed (sr);
	srunner_free (sr);

	return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

